/*
	$Date: 1996/11/04 13:45:40 $
	$Revision: 1.3 $
*/
#include <stdio.h>
#include "phrase.h"

int allocated = 0;

#ifdef USEDBM
#include <ndbm.h>
#include <sys/file.h>
#endif

unsigned char *pprefix[MAXPHRASES];
unsigned char *pcand[MAXPHRASES];
unsigned char *pcandyomi[MAXPHRASES];
int nphrases;

unsigned char *wcand[MAXWORDS];
unsigned char *wcandyomi[MAXWORDS];
int nwords;

#ifdef USEDBM
#define DBMFILE "/user/masui/strings"
#ifdef BSD
#define DBMDBFILE "/user/masui/strings.dir"
#else
#define DBMPAGFILE "/user/masui/strings.pag"
#define DBMDIRFILE "/user/masui/strings.dir"
#endif
static DBM *db;

unsigned char
*salloc(char *s)
{
	int *ip;
	unsigned char *ret;
	static int count = 0;
	char hexbuf[10];
	datum key, content;

	key.dptr = s;
	key.dsize = strlen(s);
	content = dbm_fetch(db, key);
	if(content.dptr == NULL){
		ret = (unsigned char*)strdup(s);
		sprintf(hexbuf,"%x",ret);
		content.dptr = hexbuf;
		content.dsize = strlen(hexbuf)+1;
		dbm_store(db,key,content,DBM_REPLACE);
if(++count % 2000 == 0){
	fprintf(stderr,"%d strings are put into DBM.\n",count);
}
	}
	else {
		strncpy(hexbuf,content.dptr,content.dsize);
		hexbuf[content.dsize] = '\0';
		sscanf(hexbuf,"%x",&ret);
	}
	return ret;
}
#endif

#ifdef USEHASH
#define PRIME 9973
int hashpjw(unsigned char *s)
{
	unsigned char *p;
	unsigned h = 0, g;
	for(p=s; *p; p++){
		h = (h<<4) + (*p);
		if(g = h & 0xf0000000){
			h = h ^ (g >> 24);
			h = h ^ g;
		}
	}
	return h % PRIME;
}

typedef struct hash {
	unsigned char *s;
	struct hash *next;
} Hash;

Hash *hash[PRIME];


Hash*
newhash(unsigned char *s)
{
	Hash *h;
	h = (Hash*)malloc(sizeof(Hash));
	h->s = (unsigned char*)strdup(s);
	h->next = (Hash*)NULL;
allocated++;
	return h;
}

unsigned char
*salloc(char *s)
{
	int h;
	Hash *p;

	h = hashpjw(s);
	if(p = hash[h]){
		while(1){
			if(strcmp(p->s,s)==0) return p->s;
			if(p->next == NULL){
				p->next = newhash(s);
				return p->next->s;
			}
			p = p->next;
		}
	}
	else {
		hash[h] = newhash(s);
		return hash[h]->s;
	}
}
#endif

#if !defined(USEDBM) && !defined(USEHASH)
unsigned char
*salloc(char *s)
{
allocated++;
	return (unsigned char*)strdup(s);
}
#endif

readphrases()
{
	FILE *f;
	unsigned char buf[1000];
	unsigned char s1[100],s2[100],s3[100],*s;
	int i;

#ifdef USEDBM
#ifdef BSD
	unlink(DBMDBFILE);
#else
	unlink(DBMPAGFILE);
	unlink(DBMDIRFILE);
#endif
	db = dbm_open(DBMFILE,O_CREAT|O_RDWR,0644);
	if(db == NULL){
		fprintf(stderr,"Can't open %s\n",DBMFILE);
		exit(0);
	}
#endif

	nphrases = 0;
	f = fopen(PHRASEFILEE,"r");
	if(f == NULL){
		fprintf(stderr,"Can't open %s\n",PHRASEFILEE);
		exit(0);
	}
fprintf(stderr,"Reading English phrase frequency list...\n");
	for(;fgets(buf,1000,f) != NULL && nphrases<MAXPHRASES;nphrases++){
		sscanf(buf,"%s %s %s",s1,s2,s3);
		for(s=s1;*s;s++){ if(*s == '/') *s = ' '; }
		pprefix[nphrases] = salloc(s1);
		pcand[nphrases] = salloc(s2);
		pcandyomi[nphrases] = salloc(s3);
	}
	fclose(f);
	f = fopen(PHRASEFILE,"r");
	if(f == NULL){
		fprintf(stderr,"Can't open %s\n",PHRASEFILE);
		exit(0);
	}
fprintf(stderr,"Reading Japanese phrase frequency list...\n");
	for(;fgets(buf,1000,f) != NULL && nphrases<MAXPHRASES;nphrases++){
		sscanf(buf,"%s %s %s",s1,s2,s3);
		pprefix[nphrases] = salloc(s1);
		pcand[nphrases] = salloc(s2);
		pcandyomi[nphrases] = salloc(s3);
	}
	fclose(f);

	nwords = 0;
	f = fopen(WORDFILEE,"r");
	if(f == NULL){
		fprintf(stderr,"Can't open %s\n",WORDFILEE);
		exit(0);
	}
fprintf(stderr,"Reading English word frequency list...\n");
	for(;fgets(buf,1000,f) != NULL && nwords<MAXWORDS;nwords++){
		if(buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1] = '\0';
		s = (unsigned char*)index(buf,' ');
		if(!s) continue;
		*s = '\0';
		wcand[nwords] = salloc(s+1);
		wcandyomi[nwords] = salloc(buf);
	}
	fclose(f);
	f = fopen(WORDFILE,"r");
	if(f == NULL){
		fprintf(stderr,"Can't open %s\n",WORDFILE);
		exit(0);
	}
fprintf(stderr,"Reading Japanese word frequency list...\n");
	for(;fgets(buf,1000,f) != NULL && nwords<MAXWORDS;nwords++){
		sscanf(buf,"%s %s",s1,s2);
		wcand[nwords] = salloc(s2);
		wcandyomi[nwords] = salloc(s1);
	}
	fclose(f);

fprintf(stderr,"%d phrases/%d words read\n",nphrases,nwords);

	/*
		ٽ¤٤
	*/
	for(i=0;i<nphrases/2;i++){
		unsigned char *s;
		s = pprefix[i]; pprefix[i] = pprefix[nphrases-i-1]; pprefix[nphrases-i-1] = s;
		s = pcand[i]; pcand[i] = pcand[nphrases-i-1]; pcand[nphrases-i-1] = s;
		s = pcandyomi[i]; pcandyomi[i] = pcandyomi[nphrases-i-1]; pcandyomi[nphrases-i-1] = s;
	}
	for(i=0;i<nwords/2;i++){
		unsigned char *s;
		s = wcand[i]; wcand[i] = wcand[nwords-i-1]; wcand[nwords-i-1] = s;
		s = wcandyomi[i]; wcandyomi[i] = wcandyomi[nwords-i-1]; wcandyomi[nwords-i-1] = s;
	}
printf("allocated = %d\n",allocated);
#ifdef USEDBM
#ifdef BSD
	unlink(DBMDBFILE);
#else
	unlink(DBMPAGFILE);
	unlink(DBMDIRFILE);
#endif
#endif
}

unsigned char*
addphrase(unsigned char *p, unsigned char *c, unsigned char *y)
/*
	([ץե], , ɤ)Ȥɲä롣
*/
{
	int i,j;
	unsigned char *prefix, *str, *yomi, *ret;

/* printf("addphrase(%s,%s,%s)\n",p,c,y); */

	ret = NULL;
	if(p && strlen(p) > 0 && strlen(c) > 0){
		/* Ʊȥ꤬Х١ */
		for(i=nphrases-1;i>=0;i--){
			if(strcmp(p,pprefix[i])==0 &&
			   strcmp(c,pcand[i])==0 &&
			   strcmp(y,pcandyomi[i])==0){
				prefix = pprefix[i];
				str = pcand[i];
				yomi = pcandyomi[i];
				for(;i<nphrases-2;i++){
					pprefix[i] = pprefix[i+1];
					pcand[i] = pcand[i+1];
					pcandyomi[i] = pcandyomi[i+1];
				}
				pprefix[i] = prefix;
				pcand[i] = ret = str;
				pcandyomi[i] = yomi;
				break;
			}
		}
		if(i<0 && nphrases < MAXPHRASES){
			pprefix[nphrases] = salloc(p);
			pcand[nphrases] = ret = salloc(c);
			pcandyomi[nphrases] = salloc(y);
			nphrases++;
		}
	}
	if(c && strlen(c) > 0){
		/* Ʊȥ꤬Х١ */
		for(i=nwords-1;i>=0;i--){
			if(strcmp(c,wcand[i])==0 &&
			   strcmp(y,wcandyomi[i])==0){
				str = wcand[i];
				yomi = wcandyomi[i];
				for(;i<nwords-2;i++){
					wcand[i] = wcand[i+1];
					wcandyomi[i] = wcandyomi[i+1];
				}
				wcand[i] = str;
				if(! ret) ret = str;
				wcandyomi[i] = yomi;
				break;
			}
		}
		if(i < 0 && nwords < MAXWORDS){
			wcand[nwords] = salloc(c);
			if(! ret) ret = wcand[nwords];
			wcandyomi[nwords] = salloc(y);
			nwords++;
		}
	}
/* printf("ret = %s\n",ret); */
	return ret;
}
